<?php
/**
* @version $Id: cbpaidsubscriptions.googlecheckout.php 1581 2012-12-24 02:36:44Z beat $
* @package CBSubs (TM) Community Builder Plugin for Paid Subscriptions (TM)
* @subpackage Plugin for Paid Subscriptions
* @copyright (C) 2007-2016 and Trademark of Lightning MultiCom SA, Switzerland - www.joomlapolis.com - and its licensors, all rights reserved
* @license http://www.gnu.org/licenses/old-licenses/gpl-2.0.html GNU/GPL version 2
*/

use CBLib\Xml\SimpleXMLElement;

/** Ensure this file is being included by a parent file */
if ( ! ( defined( '_VALID_CB' ) || defined( '_JEXEC' ) || defined( '_VALID_MOS' ) ) ) { die( 'Direct Access to this location is not allowed.' ); }

global $_CB_framework;

// Avoids errors in CB plugin edit:
/** @noinspection PhpIncludeInspection */
include_once( $_CB_framework->getCfg( 'absolute_path' ) . '/components/com_comprofiler/plugin/user/plug_cbpaidsubscriptions/cbpaidsubscriptions.class.php' );

// This gateway implements a payment handler using a hosted page at the PSP:
// Import class cbpaidHostedPagePayHandler that extends cbpaidPayHandler
// and implements all gateway-generic CBSubs methods.

/**
 * Payment handler class for this gateway: Handles all payment events and notifications, called by the parent class:
 *
 * OEM base
 * Please note that except the constructor and the API version this class does not implement any public methods.
 */
class cbpaidgooglecheckoutoem extends cbpaidHostedPagePayHandler
{
	/**
	 * Gateway API version used
	 * @var int
	 */
	public $gatewayApiVersion	=	"1.3.0";

	/**
	 * Constructor
	 *
	 * @param cbpaidGatewayAccount $account
	 */
	public function __construct( $account )
	{
		parent::__construct( $account );

		// Set gateway URLS for $this->pspUrl() results: first 2 are the main hosted payment page posting URL, next ones are gateway-specific:
		$this->_gatewayUrls	=	array(	'psp+normal'	=> $this->getAccountParam( 'psp_normal_url' ) . $this->getAccountParam( 'pspid' ),
										'psp+test'		=> $this->getAccountParam( 'psp_test_url' ) . $this->getAccountParam( 'pspid' ) );
	}

	/**
	 * CBSUBS HOSTED PAGE PAYMENT API METHODS:
	 */

	/**
	 * Returns single payment request parameters for gateway depending on basket (without specifying payment type)
	 *
	 * @param  cbpaidPaymentBasket  $paymentBasket   paymentBasket object
	 * @return array                                 Returns array $requestParams
	 */
	protected function getSinglePaymentRequstParams( $paymentBasket )
	{
		return array(	'amount' => sprintf( '%0.2f', $paymentBasket->mc_gross ),
						'currency_code' => $paymentBasket->mc_currency,
						'custom' => $paymentBasket->id
					);
	}

	/**
	 * Optional function: only needed for recurring payments:
	 * Returns subscription request parameters for gateway depending on basket (without specifying payment type)
	 *
	 * @param  cbpaidPaymentBasket  $paymentBasket   paymentBasket object
	 * @return array                                 Returns array $requestParams
	 */
	protected function getSubscriptionRequstParams( $paymentBasket )
	{
		// Run period conversion so errors report on basket:
		$this->_periodsLimits( explode( ' ', $paymentBasket->period3 ) );

		if ( $paymentBasket->period1 ) {
			$this->_periodsLimits( explode( ' ', $paymentBasket->period1 ) );
		}

		return $this->getSinglePaymentRequstParams( $paymentBasket );
	}

	/**
	* Handles the gateway-specific result of payments (redirects back to this site and gateway notifications). WARNING: unchecked access !
	*
	* @param  cbpaidPaymentBasket  $paymentBasket  New empty object. returning: includes the id of the payment basket of this callback (strictly verified, otherwise untouched)
	* @param  array                $postdata       _POST data for saving edited tab content as generated with getEditTab
	* @param  string               $result         result= get parameter, other than 'notify', 'success' or 'cancel'.
	* @return string                               HTML to display if frontend, text to return to gateway if notification, FALSE if registration cancelled and ErrorMSG generated, or NULL if nothing to display
	*/
	protected function handleOtherResult( $paymentBasket, $postdata, $result )
	{
		$return												=	null;

		if ( $result == 'payment' ) {
			$paymentBasketId								=	(int) cbGetParam( $postdata, 'custom', 0 );
			$exists											=	$paymentBasket->load( (int) $paymentBasketId );

			if ( $exists && ( sprintf( '%.2f', $paymentBasket->mc_gross ) == cbGetParam( $postdata, 'amount' ) ) && ( $paymentBasket->mc_currency == cbGetParam( $postdata, 'currency_code' ) ) ) {
				$rate										=	$paymentBasket->mc_gross;

				if ( $paymentBasket->period3 ) {
					list( /* $duration1 */, $type1, /*$start */ )	=	$this->_periodsLimits( explode( ' ', $paymentBasket->period3 ) );

					$type									=	'subscription';

					if ( $paymentBasket->period1 ) {
						list( /* $duration2 */, $type2, $start )	=	$this->_periodsLimits( explode( ' ', $paymentBasket->period1 ) );

						if ( $type1 == $type2 ) {
							$start							=	null;
						}

						$rate								=	$paymentBasket->mc_amount1;
					} else {
						$start								=	null;
					}
				} else {
					$type									=	'purchase';
					$type1									=	null;
					$start									=	null;
				}

				$request									=	'<?xml version="1.0" encoding="UTF-8"?>'
															.	'<checkout-shopping-cart xmlns="http://checkout.google.com/schema/2">'
															.		'<shopping-cart>'
															.			'<items>'
															.				'<item>'
															.					'<merchant-item-id>' . htmlspecialchars( $paymentBasket->item_number ) . '</merchant-item-id>'
															.					'<item-name>' . htmlspecialchars( $paymentBasket->item_name ) . '</item-name>'
															.					'<item-description></item-description>'
															.					'<unit-price currency="' . $paymentBasket->mc_currency . '">' . sprintf( '%.2f', $rate ) . '</unit-price>'
															.					'<quantity>' . sprintf( '%.0f', $paymentBasket->quantity ) . '</quantity>'
															.					'<merchant-private-item-data>'
															.						'<basket-id>' . $paymentBasket->id . '</basket-id>'
															.						'<basket-currency>' . $paymentBasket->mc_currency . '</basket-currency>'
															.						'<basket-type>' . $type . '</basket-type>'
															.					'</merchant-private-item-data>'
															.					'<digital-content>'
															.						'<display-disposition>OPTIMISTIC</display-disposition>'
															.						'<description>' . htmlspecialchars( CBPTXT::T( 'Your transaction has been completed and a receipt for your purchase has been emailed to you by Google.' ) ) . '</description>'
															.					'</digital-content>'
															.				'</item>';

				if ( $type == 'subscription' ) {
					$request								.=				'<item>'
															.					'<merchant-item-id>' . htmlspecialchars( $paymentBasket->item_number ) . '-SUB</merchant-item-id>'
															.					'<item-name>' . htmlspecialchars( $paymentBasket->item_name ) . '</item-name>'
															.					'<item-description></item-description>'
															.					'<unit-price currency="' . $paymentBasket->mc_currency . '">0</unit-price>'
															.					'<quantity>' . sprintf( '%.0f', $paymentBasket->quantity ) . '</quantity>'
															.					'<merchant-private-item-data>'
															.						'<basket-id>' . $paymentBasket->id . '</basket-id>'
															.						'<basket-currency>' . $paymentBasket->mc_currency . '</basket-currency>'
															.						'<basket-type>subscription</basket-type>'
															.					'</merchant-private-item-data>'
															.					'<subscription type="google" period="' . $type1 . '"' . ( $start ? ' start-date="' . gmdate( 'c', $start ) : null ) . '>'
															.						'<payments>'
															.							'<subscription-payment' . ( $paymentBasket->recur_times ? ' times="' . $paymentBasket->recur_times . '"' : null ) . '>'
															.								'<maximum-charge currency="' . $paymentBasket->mc_currency . '">' . sprintf( '%.2f', $paymentBasket->mc_amount3 ) . '</maximum-charge>'
															.							'</subscription-payment>'
															.						'</payments>'
															.						'<recurrent-item>'
															.							'<merchant-item-id>' . htmlspecialchars( $paymentBasket->item_number ) . '</merchant-item-id>'
															.							'<item-name>' . htmlspecialchars( $paymentBasket->item_name ) . '</item-name>'
															.							'<item-description></item-description>'
															.							'<unit-price currency="' . $paymentBasket->mc_currency . '">' . sprintf( '%.2f', $paymentBasket->mc_amount3 ) . '</unit-price>'
															.							'<quantity>' . sprintf( '%.0f', $paymentBasket->quantity ) . '</quantity>'
															.							'<merchant-private-item-data>'
															.								'<basket-id>' . $paymentBasket->id . '</basket-id>'
															.								'<basket-currency>' . $paymentBasket->mc_currency . '</basket-currency>'
															.								'<basket-type>subscription</basket-type>'
															.							'</merchant-private-item-data>'
															.						'</recurrent-item>'
															.					'</subscription>'
															.					'<digital-content>'
															.						'<display-disposition>OPTIMISTIC</display-disposition>'
															.						'<description>' . htmlspecialchars( CBPTXT::T( 'Your transaction has been completed and a receipt for your purchase has been emailed to you by Google.' ) ) . '</description>'
															.					'</digital-content>'
															.				'</item>';
				}

				$request									.=			'</items>'
															.		'</shopping-cart>'
															.		'<checkout-flow-support>'
															.			'<merchant-checkout-flow-support>'
															.				'<edit-cart-url>' . htmlspecialchars( $this->getCancelUrl( $paymentBasket ) ) . '</edit-cart-url>'
															.				'<continue-shopping-url>' . htmlspecialchars( $this->getSuccessUrl( $paymentBasket ) ) . '</continue-shopping-url>'
															.			'</merchant-checkout-flow-support>'
															.		'</checkout-flow-support>'
															.	'</checkout-shopping-cart>';

				$response									=	null;
				$status										=	null;
				$error										=	$this->_httpsRequest( str_replace( 'checkoutForm', 'merchantCheckout', $this->gatewayUrl( 'psp' ) ), $request, 45, $response, $status, 'post', 'xml', '*/*', true, 443, $this->getAccountParam( 'pspid' ), $this->getAccountParam( 'pspkey' ) );

				$requestdata								=	null;

				if ( $response ) {
					$xmlresponse							=	new SimpleXMLElement( $response );

					if ( $xmlresponse ) {
						$requestdata						=	$this->xmlTagValuesToArray( $xmlresponse );
					}
				}

				if ( ( ! $error ) && ( $status == 200 ) && $requestdata ) {
					$redirect								=	cbGetParam( $requestdata, 'redirect-url' );

					if ( $redirect ) {
						cbRedirect( $redirect );
					} else {
						$this->_setLogErrorMSG( 3, null, $this->getPayName() . ': merchantCheckout responded without a redirect url! Error: ' . cbGetParam( $requestdata, 'error-message' ), CBPTXT::T( 'Checkout failed.' ) . ' ' . CBPTXT::T( 'Please contact site administrator to check error log.' ) );

						$return								=	false;
					}
				} else {
					if ( $requestdata ) {
						$this->_setLogErrorMSG( 3, null, $this->getPayName() . ': merchantCheckout failed to respond! Error: ' . cbGetParam( $requestdata, 'error-message' ), CBPTXT::T( 'Checkout failed.' ) . ' ' . CBPTXT::T( 'Please contact site administrator to check error log.' ) );
					} else {
						$this->_setLogErrorMSG( 3, null, $this->getPayName() . ': merchantCheckout failed to respond. Please check Merchant ID and Merchant Key settings', CBPTXT::T( 'Checkout failed.' ) . ' ' . CBPTXT::T( 'Please contact site administrator to check error log.' ) );
					}

					$return									=	false;
				}
			} else {
				$this->_setErrorMSG( CBPTXT::T( 'Payment basket does not login.' ) );

				$return										=	false;
			}
		}

		return $return;
	}


	/**
	 * The user got redirected back from the payment service provider with a success message: Let's see how successfull it was
	 *
	 * @param  cbpaidPaymentBasket  $paymentBasket  New empty object. returning: includes the id of the payment basket of this callback (strictly verified, otherwise untouched)
	 * @param  array                $postdata       _POST data for saving edited tab content as generated with getEditTab
	 * @return string                               HTML to display if frontend, FALSE if XML error (and not yet ErrorMSG generated), or NULL if nothing to display
	 */
	protected function handleReturn( $paymentBasket, $postdata )
	{
		// Google has no PDT so lets just add the basket id from GET parameters were we've passed the success url as the return url:
		$requestdata	=	array( 'basket-id' => (int) cbGetParam( $_GET, 'cbpbasket', null ) );

		return $this->_returnParamsHandler( $paymentBasket, $requestdata, 'R' );
	}

	/**
	 * The user cancelled his payment
	 *
	 * @param  cbpaidPaymentBasket  $paymentBasket  New empty object. returning: includes the id of the payment basket of this callback (strictly verified, otherwise untouched)
	 * @param  array                $postdata       _POST data for saving edited tab content as generated with getEditTab
	 * @return string                               HTML to display, FALSE if registration cancelled and ErrorMSG generated, or NULL if nothing to display
	 */
	protected function handleCancel( $paymentBasket, $postdata )
	{
		// The user cancelled his payment (and registration):
		if ( $this->hashPdtBackCheck( $this->_getReqParam( 'pdtback' ) ) ) {
			$paymentBasketId					=	(int) $this->_getReqParam( 'basket' );

			// check if cancel was from gateway:
			if ( ! $paymentBasketId ) {
				$paymentBasketId				=	(int) $this->_getReqParam( 'cbpbasket' );
			}

			$exists								=	$paymentBasket->load( (int) $paymentBasketId );

			if ( $exists && ( $this->_getReqParam( 'id' ) == $paymentBasket->shared_secret ) && ( $paymentBasket->payment_status != 'Completed' ) ) {
				$paymentBasket->payment_status	=	'RedisplayOriginalBasket';

				$this->_setErrorMSG( CBPTXT::T( 'Payment cancelled.' ) );
			}
		}

		return false;
	}

	/**
	 * The payment service provider server did a server-to-server notification: Verify and handle it here:
	 *
	 * @param  cbpaidPaymentBasket  $paymentBasket  New empty object. returning: includes the id of the payment basket of this callback (strictly verified, otherwise untouched)
	 * @param  array                $postdata       _POST data for saving edited tab content as generated with getEditTab
	 * @return string                              Text to return to gateway if notification, or NULL if nothing to display
	 */
	protected function handleNotification( $paymentBasket, $postdata )
	{
		if ( ( count( $postdata ) > 0 ) && isset( $postdata['serial-number'] ) ) {
			// we prefer POST for sensitive data:
			$requestdata	=	$postdata;
		} else {
			// but if gateway needs GET, we will work with it too:
			$requestdata	=	$this->_getGetParams();
		}

		$this->_returnParamsHandler( $paymentBasket, $requestdata, 'I' );
	}

	/**
	* Cancels an existing recurring subscription
	*
	 * @param  cbpaidPaymentBasket  $paymentBasket  paymentBasket object
	 * @param  cbpaidPaymentItem[]  $paymentItems   redirect immediately instead of returning HTML for output
	 * @return boolean|string                       TRUE if unsubscription done successfully, STRING if error
	*/
	protected function handleStopPaymentSubscription( $paymentBasket, $paymentItems )
	{
		global $_CB_framework;

		$return									=	false;

		if ( $paymentBasket->mc_amount3 ) {
			$subscription_id					=	$paymentBasket->subscr_id;

			if ( $subscription_id ) {
				$request						=	'<?xml version="1.0" encoding="UTF-8"?>'
												.	'<cancel-items xmlns="http://checkout.google.com/schema/2" google-order-number="' . htmlspecialchars( $subscription_id ) . '">'
												.		'<reason>' . htmlspecialchars( CBPTXT::T( 'Cancel automatic payments' ) ) . '</reason>'
												.		'<item-ids>'
												.			'<item-id>'
												.				'<merchant-item-id>' . htmlspecialchars( $paymentBasket->item_number ) . '-SUB</merchant-item-id>'
												.			'</item-id>'
												.		'</item-ids>'
												.	'</cancel-items>';

				$response						=	null;
				$status							=	null;
				$resultUrl						=	str_replace( 'checkoutForm', 'request', $this->gatewayUrl( 'psp' ) );
				$error							=	$this->_httpsRequest( $resultUrl, $request, 45, $response, $status, 'post', 'xml', '*/*', true, 443, $this->getAccountParam( 'pspid' ), $this->getAccountParam( 'pspkey' ) );

				if ( $error || ( $status != 200 ) || ( ! $response ) ) {
					$this->_setLogErrorMSG( 3, null, $this->getPayName() . ': ' . sprintf( 'Failed accessing request url: %s : Got HTTP Status: %s and reply: %s', $resultUrl, $status, $response ), CBPTXT::T( 'Submitted unsubscription failed on-site.' ) . ' ' . CBPTXT::T( 'Please contact site administrator to check error log.' ) );
				} else {
					$xml						=	new SimpleXMLElement( $response );

					if ( $xml->getName() == 'request-received' ) {
						$ipn					=	$this->_prepareIpn( 'R', $paymentBasket->payment_status, $paymentBasket->payment_type, 'Unsubscribe', $_CB_framework->now(), 'utf-8' );
						$ipn->test_ipn			=	$paymentBasket->test_ipn;
						$ipn->raw_result		=	'SUCCESS';
						$ipn->raw_data			=	'$message_type="STOP_PAYMENT_SUBSCRIPTION"' . ";\n"
												.	/* cbGetParam() not needed: we want raw info */ '$xml_response=' . $response . ";\n"
												.	/* cbGetParam() not needed: we want raw info */ '$_GET=' . var_export( $_GET, true ) . ";\n"
												.	/* cbGetParam() not needed: we want raw info */ '$_POST=' . var_export( $_POST, true ) . ";\n";

						$ipn->bindBasket( $paymentBasket );

						$bskToIpn				=	array(	'sale_id' => 'sale_id',
															'txn_id' => 'txn_id',
															'subscr_id' => 'subscr_id',
															'user_id' => 'user_id'
														);

						foreach ( $bskToIpn as $k => $v ) {
							$ipn->$k			=	$paymentBasket->$v;
						}

						$ipn->txn_type			=	'subscr_cancel';

						if ( ! $ipn->store() ) {
							$this->_setLogErrorMSG( 3, null, $this->getPayName() . ': unsubscribe IPN failed to store', CBPTXT::T( 'Submitted unsubscription failed on-site.' ) . ' ' . CBPTXT::T( 'Please contact site administrator to check error log.' ) );
						} else {
							$return				=	true;
						}
					} else {
						$this->_setLogErrorMSG( 3, null, $this->getPayName() . ': ' . sprintf( 'Malformed XML reply from request url: %s . Response: %s', $resultUrl, $response ), CBPTXT::T( 'Submitted unsubscription failed on-site.' ) . ' ' . CBPTXT::T( 'Please contact site administrator to check error log.' ) );
					}
				}
			} else {
				$this->_setLogErrorMSG( 3, null, $this->getPayName() . ': unsubscribe failed from missing subscr_id in payment basket', CBPTXT::T( 'Submitted unsubscription failed on-site.' ) . ' ' . CBPTXT::T( 'Please contact site administrator to check error log.' ) );
			}
		}

		return $return;
	}

	/**
	 * Refunds a payment
	 *
	 * @param  cbpaidPaymentBasket       $paymentBasket  paymentBasket object
	 * @param  cbpaidPayment             $payment        payment object
	 * @param  cbpaidPaymentItem[]|null  $paymentItems   Array of payment items to refund completely (if no $amount)
	 * @param  boolean                   $lastRefund     Last refund for $payment ?
	 * @param  float                     $amount         Amount in currency of the payment
	 * @param  string                    $reasonText     Refund reason comment text for gateway
	 * @param  string                    $returnText     RETURN param : Text of success message or of error message
	 * @return boolean                                   true if refund done successfully, false if error
	 */
	public function refundPayment( $paymentBasket, $payment, $paymentItems, $lastRefund, $amount, $reasonText, &$returnText )
	{
		$return						=	false;

		$request					=	'<?xml version="1.0" encoding="UTF-8"?>'
									.	'<refund-order xmlns="http://checkout.google.com/schema/2" google-order-number="' . htmlspecialchars( $payment->txn_id ) . '">'
									.		'<reason>' . htmlspecialchars( ( $reasonText ? $reasonText : CBPTXT::T( 'Refund payment' ) ) ) . '</reason>'
									.		( $amount != $payment->mc_gross ? '<amount>' . sprintf( '%.2f', $amount ) . '</amount>' : null )
									.	'</refund-order>';

		$response					=	null;
		$status						=	null;
		$resultUrl					=	str_replace( 'checkoutForm', 'request', $this->gatewayUrl( 'psp' ) );
		$error						=	$this->_httpsRequest( $resultUrl, $request, 45, $response, $status, 'post', 'xml', '*/*', true, 443, $this->getAccountParam( 'pspid' ), $this->getAccountParam( 'pspkey' ) );

		if ( $error || ( $status != 200 ) || ( ! $response ) ) {
			$this->_setLogErrorMSG( 3, null, $this->getPayName() . ': ' . sprintf( 'Failed accessing request url: %s : Got HTTP Status: %s and reply: %s', $resultUrl, $status, $response ), CBPTXT::T( 'Submitted unsubscription failed on-site.' ) . ' ' . CBPTXT::T( 'Please contact site administrator to check error log.' ) );
		} else {
			$requestdata			=	array();
			$additionalLogData		=	array();
			$apiRequest				=	array(	'request' => $request,
												'response' => $response,
												'status' => $status,
												'resultUrl' => $resultUrl,
												'error' => $error,
											);

			if ( $amount != $payment->mc_gross ) {
				$type				=	'4';
			} else {
				$type				=	'3';
			}

			$return					=	$this->_returnParamsHandler( $paymentBasket, $requestdata, $type, $additionalLogData, $payment, $apiRequest );
		}

		return $return;
	}

	/**
	 * gives gateway button URL server name from gateway URL list
	 *
	 * @param  cbpaidPaymentBasket  $paymentBasket  paymentBasket object
	 * @param  boolean              $autoRecurring   TRUE: autorecurring payment, FALSE: single payment
	 * @return string  server-name (with 'https://' )
	 */
	protected function pspUrl( $paymentBasket, $autoRecurring )
	{
		return $this->pspRedirectUrl( $paymentBasket, array(), $autoRecurring );
	}

	/**
	 * Returns https redirect URL for redirections to gateway for payment with $requestParams
	 *
	 * @param  cbpaidPaymentBasket  $paymentBasket  paymentBasket object
	 * @param  array                $requestParams
	 * @param  boolean              $autoRecurring   TRUE: autorecurring payment, FALSE: single payment
	 * @return string                                Full URL for redirect including https:// in front
	 */
	protected function pspRedirectUrl( $paymentBasket, $requestParams, $autoRecurring )
	{
		return $this->cbsubsGatewayUrl( 'payment', null, $paymentBasket, $requestParams, false );
	}

	/**
	 * Compute the CBSubs payment_status based on gateway's reply in $postdata:
	 *
	 * @param  array   $postdata  raw POST data received from the payment gateway
	 * @param  string  $reason    OUT: reason_code
     * @return string             CBSubs payment status
	 */
	private function _paymentStatus( $postdata, &$reason )
	{
		$accept_payment_condition	=	$this->getAccountParam( 'accept_payment_condition' );
		$status						=	cbGetParam( $postdata, 'order-status' );

		// We treat chargeable as charged as for digital goods the charge + shipping is instant:
		switch ( strtoupper( $status ) ) {
			case 'CHARGEABLE':
			case 'CHARGING':
			case 'CHARGED':
				$reason				=	null;
				$status				=	'Completed';
				break;
			case 'REVIEWING':
				if ( $accept_payment_condition == 'pending' ) {
					$reason			=	null;
					$status			=	'Completed';
				} else {
					$reason			=	'Transaction being processed';
					$status			=	'Pending';
				}
				break;
			case 'REFUND':
				$reason				=	null;
				$status				=	'Refunded';
				break;
			case 'PAYMENT_DECLINED':
			case 'CANCELLED':
			case 'CANCELLED_BY_GOOGLE':
				$reason				=	'Payment declined';
				$status				=	'Denied';
				break;
		}

		return $status;
	}

	/**
	 * parse period into period and type then calculate new period limitations:
	 *
	 * @param array $periodTypeArray
	 * @return array
	 */
	private function _periodsLimits( $periodTypeArray )
	{
		global $_CB_framework;

		$p			=	$periodTypeArray[0];
		$t			=	$periodTypeArray[1];
		$now		=	$_CB_framework->now();

		$systemTimeZone	=	new DateTimeZone( cbpaidTimes::getInstance()->systemTimeZone() );
		$date		=	new DateTime( '@' . (string) $now, $systemTimeZone );
		$date->setTimezone( $systemTimeZone );

		if ( $t == 'Y' ) {
			$t		=	'YEARLY';

			if ( $p > 1 ) {
				trigger_error( sprintf( 'Trying to subscribe %s years to Google Checkout, subscription augmented to 1 year.', $p ), E_USER_WARNING );

				$p	=	1;
			}

			$date->modify( '+' . intval( $p ) . ' YEAR' );
		} elseif ( $t == 'M' ) {
			$t		=	'MONTHLY';

			if ( $p == 2 ) {
				$t	=	'EVERY_TWO_MONTHS';
			} elseif ( $p == 3 ) {
				$t	=	'QUARTERLY';
			} elseif ( $p == 12 ) {
				$t	=	'YEARLY';
			} elseif ( $p > 1 ) {
				trigger_error( sprintf( 'Trying to subscribe %s months to Google Checkout, subscription augmented to 1 month.', $p ), E_USER_WARNING );

				$p	=	1;
			}

			$date->modify( '+' . intval( $p ) . ' MONTH' );
		} elseif  ( $t == 'W' ) {
			$t		=	'WEEKLY';

			if ( $p > 1 ) {
				trigger_error( sprintf( 'Trying to subscribe %s weeks to Google Checkout, subscription augmented to 1 week.', $p ), E_USER_WARNING );

				$p	=	1;
			}

			$date->modify( '+' . intval( $p ) . ' WEEK' );
		} elseif ( $t == 'D' ) {
			$t		=	'DAILY';

			if ( $p == 7 ) {
				$t	=	'WEEKLY';
			} elseif ( $p == 15 ) {
				$t	=	'SEMI_MONTHLY';
			} elseif ( $p == 30 ) {
				$t	=	'MONTHLY';
			} elseif ( $p == 90 ) {
				$t	=	'QUARTERLY';
			} elseif ( $p == 365 ) {
				$t	=	'YEARLY';
			} elseif ( $p > 1 ) {
				trigger_error( sprintf( 'Trying to subscribe %s days to Google Checkout, subscription augmented to 1 day.', $p ), E_USER_WARNING );

				$p	=	1;
			}

			$date->modify( '+' . intval( $p ) . ' DAY' );
		} else {
			trigger_error( __CLASS__ . '::' . __FUNCTION__ . ' wrong parameters.', E_USER_WARNING );

			$date->modify( '@0' );
		}

		return array( $p, $t, $date->getTimestamp() );
	}

	/**
	 * The user got redirected back from the payment service provider with a success message: let's see how successfull it was
	 *
	 * @param  cbpaidPaymentBasket  $paymentBasket       New empty object. returning: includes the id of the payment basket of this callback (strictly verified, otherwise untouched)
	 * @param  array                $requestdata         Data returned by gateway
	 * @param  string               $type                Type of return ('R' for PDT, 'I' for INS, 'A' for Autorecurring payment)
     * @param  array                $additionalLogData   Additional strings to log with IPN
	 * @param  cbpaidPayment        $payment             (optional): Needed only for refunds ($type '3' or '4')
	 * @param  array                $apiRequest          (optional): Needed only for refunds ($type '3' or '4')
     * @return string                                    HTML to display if frontend, text to return to gateway if notification, FALSE if registration cancelled and ErrorMSG generated, or NULL if nothing to display
	 */
	private function _returnParamsHandler( $paymentBasket, $requestdata, $type, $additionalLogData = null, $payment = null, $apiRequest = null )
	{
		global $_CB_framework, $_GET, $_POST;

		$ret														=	null;
		$error														=	null;
		$response													=	null;
		$resultUrl													=	null;

		// IPN doesn't return transacton information; lets request for it:
		$serial_number												=	cbGetParam( $requestdata, 'serial-number', null );

		// IPN should only continue if serial number is present. PDT contains the basket id already as is from success url:
		if ( ( $serial_number && ( $type == 'I' ) ) || in_array( $type, array( '3', '4' ) ) || ( $type == 'R' ) ) {
			// IPN only gives us the serial number; request rest of notification:
			if ( in_array( $type, array( 'I', '3', '4' ) ) ) {
				// Request notification information:
				if ( in_array( $type, array( '3', '4' ) ) && $apiRequest ) {
					$request										=	$apiRequest['request'];
					$response										=	$apiRequest['response'];
					$status											=	$apiRequest['status'];
					$resultUrl										=	$apiRequest['resultUrl'];
					$error											=	$apiRequest['error'];
				} else {
					// Let Google know the IPN was received; without this Google will resend constantly for up to 15 days:
					$acknowledgment									=	'<?xml version="1.0" encoding="UTF-8"?>'
																	.	'<notification-acknowledgment xmlns="http://checkout.google.com/schema/2" serial-number="' . htmlspecialchars( $serial_number ) . '" />';

					echo $acknowledgment;

					$request										=	'<?xml version="1.0" encoding="UTF-8"?>'
																	.	'<notification-history-request xmlns="http://checkout.google.com/schema/2">'
																	.		'<serial-number>' . htmlspecialchars( $serial_number ) . '</serial-number>'
																	.	'</notification-history-request>';

					$response										=	null;
					$status											=	null;
					$resultUrl										=	str_replace( 'checkoutForm', 'reports', $this->gatewayUrl( 'psp' ) );
					$error											=	$this->_httpsRequest( $resultUrl, $request, 45, $response, $status, 'post', 'xml', '*/*', true, 443, $this->getAccountParam( 'pspid' ), $this->getAccountParam( 'pspkey' ) );
				}

				if ( ( ! $error ) && ( $status == 200 ) && $response ) {
					$xml											=	new SimpleXMLElement( $response );

					if ( $xml->getName() == 'request-received' ) {
						$xml										=	$xml->children();
						$xml_response								=	$this->xmlTagValuesToArray( $xml->children() );
					} else {
						$xml_response								=	$this->xmlTagValuesToArray( $xml );
					}

					if ( $xml_response ) {
						$requestdata								=	$xml_response;

						// We need to push basket id and currency to root for easier usage as its deep inside the items custom data:
						$requestdata['basket-id']					=	(string) $xml->getElementByPath( 'order-summary/shopping-cart/items/item/merchant-private-item-data/basket-id' ); // Needed to load the basket
						$requestdata['basket-currency']				=	(string) $xml->getElementByPath( 'order-summary/shopping-cart/items/item/merchant-private-item-data/basket-currency' ); // Needed for validation checks
						$requestdata['basket-type']					=	(string) $xml->getElementByPath( 'order-summary/shopping-cart/items/item/merchant-private-item-data/basket-type' ); // Needed to determine type of IPN
						$requestdata['order-total']					=	(string) $xml->getElementByPath( 'order-summary/order-total' ); // Needed to determine gateway reported total of the order
						$requestdata['order-status']				=	(string) $xml->getElementByPath( 'order-summary/financial-order-state' ); // Needed to determine the current status of the order

						// Google doesn't have a refund/chargeback status so lets figure one out and pass the refunded amount:
						if ( in_array( (string) $xml->getName(), array( 'refund-amount-notification', 'chargeback-amount-notification' ) ) ) {
							$refund_ammount							=	(string) $xml->getElementByPath( 'total-refund-amount' );
							$chareback_ammount						=	(string) $xml->getElementByPath( 'total-chargeback-amount' );

							if ( $refund_ammount || $chareback_ammount ) {
								$requestdata['order-status']		=	'REFUND';

								if ( $refund_ammount ) {
									$requestdata['refund-total']	=	$refund_ammount;
								} elseif ( $chareback_ammount ) {
									$requestdata['refund-total']	=	$chareback_ammount;
								}
							}
						}
					} else {
						$error										=	sprintf( 'Malformed XML reply from notification check: Got response %s for request %s . Using Requestdata: %s', $response, $resultUrl, var_export( $requestdata, true ) );
					}
				} else {
					$error											=	sprintf( 'Failed accessing notification url: %s : Got HTTP Status: %s and reply: %s . Using Requestdata: %s', $resultUrl, $status, $response, var_export( $requestdata, true ) );
				}
			}

			$exists													=	false;

			if ( ! $error ) {
				$paymentBasketId									=	(int) cbGetParam( $requestdata, 'basket-id', null );

				if ( $paymentBasketId ) {
					$exists											=	$paymentBasket->load( (int) $paymentBasketId );

					if ( ! $exists ) {
						if ( in_array( $type, array( 'I', '3', '4' ) ) ) {
							$error									=	sprintf( 'Payment basket %s (basket-id) does not exist for transaction result %', $paymentBasketId, $response );
						}
					}
				} else {
					if ( in_array( $type, array( 'I', '3', '4' ) ) ) {
						$error										=	sprintf( 'basket-id is missing in the notification result: %s . Parsed result: %s of request: %s', $response, var_export( $requestdata, true ), $resultUrl );
					} else {
						$error										=	'basket-id is missing in the return URL: ' . var_export( $_GET, true );
					}
				}
			} else {
				$paymentBasketId									=	null;
			}

			if ( ! $error ) {
				if ( $exists && ( ( cbGetParam( $requestdata, $this->_getPagingParamName( 'id' ), 0 ) == $paymentBasket->shared_secret ) && ( ! ( ( ( $type == 'R' ) || ( $type == 'I' ) ) && ( ( $paymentBasket->payment_status == 'Completed' ) && ( ! ( ( $type == 'I' ) && ( cbGetParam( $requestdata, 'order-status' ) != 'REFUND' ) ) ) ) ) ) ) ) {
					// Log the return record:
					$log_type										=	$type;
					$reason											=	null;
					$paymentStatus									=	$this->_paymentStatus( $requestdata, $reason );
					$paymentType									=	'Credit Card';
					$paymentTime									=	$_CB_framework->now();

					if ( $paymentStatus == 'Error' ) {
						$errorTypes									=	array( 'I' => 'D', 'R' => 'E', '3' => 'V', '4' => 'V' );

						if ( isset( $errorTypes[$type] ) ) {
							$log_type								=	$errorTypes[$type];
						}
					}

					$ipn											=	$this->_prepareIpn( $log_type, $paymentStatus, $paymentType, $reason, $paymentTime, 'utf-8' );

					if ( $paymentStatus == 'Refunded' ) {
						// in case of refund we need to log the payment as it has same TnxId as first payment: so we need payment_date for discrimination:
						$ipn->payment_date							=	gmdate( 'H:i:s M d, Y T', $paymentTime ); // paypal-style
					}

					$ipn->test_ipn									=	( $this->getAccountParam( 'normal_gateway' ) == '0' ? 1 : 0 );		// Google has no param telling if it is for test or real.
					$ipn->raw_data									=	'$message_type="' . ( $type == 'R' ? 'RETURN_TO_SITE' : ( $type == 'I' ? 'NOTIFICATION' : ( $type == '3' ? 'REFUND' : ( $type == '4' ? 'PARTIAL_REFUND' : 'UNKNOWN' ) ) ) ) . '";' . "\n";

					if ( $additionalLogData ) {
						foreach ( $additionalLogData as $k => $v ) {
							$ipn->raw_data							.=	'$' . $k . '="' . var_export( $v, true ) . '";' . "\n";
						}
					}

					$ipn->raw_data									.=	/* cbGetParam() not needed: we want raw info */ '$requestdata=' . var_export( $requestdata, true ) . ";\n"
																	.	/* cbGetParam() not needed: we want raw info */ '$_GET=' . var_export( $_GET, true ) . ";\n"
																	.	/* cbGetParam() not needed: we want raw info */ '$_POST=' . var_export( $_POST, true ) . ";\n";

					if ( $paymentStatus == 'Error' ) {
						$paymentBasket->reason_code					=	$reason;

						$this->_storeIpnResult( $ipn, 'ERROR:' . $reason );
						$this->_setLogErrorMSG( 4, $ipn, $this->getPayName() . ': ' . $reason, CBPTXT::T( 'Sorry, the payment server replied with an error.' ) . ' ' . CBPTXT::T( 'Please contact site administrator to check payment status and error log.' ) );

						$ret										=	false;
					} else {
						$ipn->bindBasket( $paymentBasket );

						$ipn->sale_id								=	$paymentBasketId;

						$order_summary								=	cbGetParam( $requestdata, 'order-summary' );

						if ( $order_summary ) {
							$insToIpn								=	array(	'txn_id' => 'google-order-number',
																				'payer_id' => 'buyer-id'
																			);

							foreach ( $insToIpn as $k => $v ) {
								$ipn->$k							=	cbGetParam( $order_summary, $v );
							}

							$shipping_address						=	cbGetParam( $order_summary, 'buyer-shipping-address' );

							if ( $shipping_address ) {
								$insToIpn							=	array(	'address_street' => 'address1',
																				'address_zip' => 'postal-code',
																				'address_city' => 'city',
																				'address_country' => 'country-code',
																				'address_state' => 'region',
																				'payer_email' => 'email',
																				'contact_phone' => 'phone'
																			);

								foreach ( $insToIpn as $k => $v ) {
									$ipn->$k						=	cbGetParam( $shipping_address, $v );
								}

								$name								=	cbGetParam( $shipping_address, 'contact-name' );

								if ( $name ) {
									$contact_name					=	explode( ' ', $name, 2 );

									if ( count( $contact_name ) < 2 ) {
										$contact_name				=	array( $name, '' );
									}

									$ipn->first_name				=	$contact_name[0];
									$ipn->last_name					=	$contact_name[1];
								}
							}
						}

						$ipn->mc_gross								=	sprintf( '%.2f', cbGetParam( $requestdata, 'order-total' ) );

						if ( ( $paymentStatus == 'Refunded' ) && ( $ipn->mc_gross > 0 ) ) {
							$ipn->mc_gross							=	( - sprintf( '%.2f', cbGetParam( $requestdata, 'refund-total' ) ) );

							if ( $payment ) {
								$ipn->computeRefundedTax( $payment );

								$ipn->parent_txn_id					=	$payment->txn_id;
							}
						}

						$ipn->mc_currency							=	cbGetParam( $requestdata, 'basket-currency' );
						$ipn->user_id								=	(int) $paymentBasket->user_id;

						$recurring									=	( cbGetParam( $requestdata, 'basket-type' ) == 'subscription' ? true : false );

						if ( $recurring ) {
							if ( ( $paymentStatus == 'Completed' ) && ( ! $paymentBasket->subscr_id ) && ( $type != 'A' ) ) {
								$ipn->txn_type						=	'subscr_signup';
								$ipn->subscr_id						=	$ipn->txn_id;
								$ipn->subscr_date					=	$ipn->payment_date;
							} elseif ( $paymentStatus == 'Denied' ) {
								if ( ( $paymentBasket->reattempts_tried + 1 ) <= cbpaidScheduler::getInstance( $this )->retries ) {
									$ipn->txn_type					=	'subscr_failed';
								} else {
									$ipn->txn_type					=	'subscr_cancel';
								}
							} elseif ( in_array( $paymentStatus, array( 'Completed', 'Processed', 'Pending' ) ) ) {
								$ipn->txn_type						=	'subscr_payment';
							}
						} else {
							$ipn->txn_type							=	'web_accept';
						}

						if ( ( $paymentBasketId == cbGetParam( $requestdata, 'basket-id', null ) ) && ( ( sprintf( '%.2f', $paymentBasket->mc_gross ) == $ipn->mc_gross ) || ( $ipn->payment_status == 'Refunded' ) ) && ( $paymentBasket->mc_currency == $ipn->mc_currency ) ) {
							if ( in_array( $ipn->payment_status, array( 'Completed', 'Processed', 'Pending', 'Refunded' ) ) ) {
								if ( ( $type == '4' ) || ( ( $ipn->payment_status == 'Refunded' ) && ( sprintf( '%.2f', $paymentBasket->mc_gross ) != sprintf( '%.2f', - $ipn->mc_gross ) ) ) ) {
									$ipn->payment_status			=	'Partially-Refunded';
								}

								$this->_storeIpnResult( $ipn, 'SUCCESS' );
								$this->_bindIpnToBasket( $ipn, $paymentBasket );

								// add the gateway to the basket:
								$paymentBasket->payment_method		=	$this->getPayName();
								$paymentBasket->gateway_account		=	$this->getAccountParam( 'id' );

								// 0: not auto-recurring, 1: auto-recurring without payment processor notifications, 2: auto-renewing with processor notifications updating $expiry_date:
								$autorecurring_type					=	( in_array( $ipn->txn_type, array( 'subscr_payment', 'subscr_signup', 'subscr_modify', 'subscr_eot', 'subscr_cancel', 'subscr_failed' ) ) ? 2 : 0 );

								// 0: not auto-renewing (manual renewals), 1: asked for by user, 2: mandatory by configuration:
								$autorenew_type						=	( $autorecurring_type ? ( ( $this->getAccountParam( 'enabled', 0 ) == 3 ) && ( $paymentBasket->isAnyAutoRecurring() == 2 ) ? 1 : 2 ) : 0 );

								if ( $recurring ) {
									$paymentBasket->reattempt		=	1; // we want to reattempt auto-recurring payment in case of failure
								}

								$this->updatePaymentStatus( $paymentBasket, $ipn->txn_type, $ipn->payment_status, $ipn, 1, $autorecurring_type, $autorenew_type, false );

								if ( in_array( $ipn->payment_status, array( 'Completed', 'Processed', 'Pending' ) ) || ( in_array( $type, array( '3', '4' ) ) && in_array( $ipn->payment_status, array( 'Refunded', 'Partially-Refunded' ) ) ) ) {
									$ret							=	true;
								}
							} else {
								// Status = 'Denied' or other:
								$this->_storeIpnResult( $ipn, ( $ipn->payment_status == 'Denied' ? 'SUCCESS' :'UKNOWN' ) );

								$paymentBasket->payment_status		=	$ipn->payment_status;

								$this->_setErrorMSG( '<div class="alert alert-info">' . $this->getTxtNextStep( $paymentBasket ) . '</div>' );

								$paymentBasket->payment_status		=	'RedisplayOriginalBasket';
								$ret								=	false;
							}
						} else {
							if ( ( $type == 'R' ) && ( $paymentBasket->payment_status == 'Pending' ) ) {
								$ret								=	true;
							} else {
								$this->_storeIpnResult( $ipn, 'MISMATCH' );
								$this->_setLogErrorMSG( 3, $ipn, $this->getPayName() . ': amount or currency missmatch or IPN failed before return', CBPTXT::T( 'Sorry, the payment does not match the basket.' ) . ' ' . CBPTXT::T( 'Please contact site administrator to check error log.' ) );

								$ret								=	false;
							}
						}
					}
				}
			} else {
				$this->_setLogErrorMSG( 3, null, $this->getPayName() . ': ' . $error , CBPTXT::T( 'Please contact site administrator to check error log.' ) );
			}
		}

		return $ret;
	}
}

/**
 * Payment account class for this gateway: Stores the settings for that gateway instance, and is used when editing and storing gateway parameters in the backend.
 *
 * OEM base
 * No methods need to be implemented or overriden in this class, except to implement the private-type params used specifically for this gateway:
 */
class cbpaidGatewayAccountgooglecheckoutoem extends cbpaidGatewayAccounthostedpage
{
}

/**
 * Payment handler class for this gateway: Handles all payment events and notifications, called by the parent class:
 *
 * Gateway-specific
 * Please note that except the constructor and the API version this class does not implement any public methods.
 */
class cbpaidgooglecheckout extends cbpaidgooglecheckoutoem
{
}

/**
 * Payment account class for this gateway: Stores the settings for that gateway instance, and is used when editing and storing gateway parameters in the backend.
 *
 * Gateway-specific
 * No methods need to be implemented or overriden in this class, except to implement the private-type params used specifically for this gateway:
 */
class cbpaidGatewayAccountgooglecheckout extends cbpaidGatewayAccountgooglecheckoutoem
{
}
